"""Unit test for amulet.wait"""

import unittest
import sys
import yaml

from amulet.helpers import (
    JujuVersion,
    environments,
    default_environment,
    juju,
    raise_status,
)

from mock import patch, Mock

RAW_ENVIRONMENTS_YAML = '''
default: gojuju
environments:
  gojuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: gojuju-xyxyz
    admin-secret: zyxyx
    default-series: world
  pyjuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: juju-xyxyz
    admin-secret: zyxyxy
    default-series: world
    juju-origin: ppa
    ssl-hostname-verification: true'''


class HelpersTest(unittest.TestCase):
    @patch('amulet.helpers.juju')
    def test_jujuversion_go(self, mj):
        mj.side_effect = ['1.2.3-series-xxx']
        version = JujuVersion()

        self.assertEqual(version.major, 1)
        self.assertEqual(version.minor, 2)
        self.assertEqual(version.patch, 3)
        self.assertEqual(str(version), '1.2.3')

        mj.assert_called_with(['version'])

    @patch('amulet.helpers.juju')
    def test_jujuversion_py(self, mj):
        mj.side_effect = [OSError('Non-zero exit'), 'juju 8.6']
        version = JujuVersion()

        self.assertEqual(version.major, 8)
        self.assertEqual(version.minor, 6)
        self.assertEqual(version.patch, None)
        self.assertEqual(str(version), '8.6')

        mj.assert_called_with(['--version'])

    @patch('amulet.helpers.juju')
    def test_jujuversion_malformed(self, mj):
        mj.return_value = '1.2.3.45'
        version = JujuVersion()

        self.assertEqual(version.major, 1)
        self.assertEqual(version.minor, 2)
        self.assertEqual(version.patch, 3)
        self.assertEqual(str(version), '1.2.3')

        mj.assert_called_once_with(['version'])

    @patch('os.path.isfile')
    @patch('builtins.open' if sys.version_info > (3,) else '__builtin__.open')
    def test_environments(self, mock_open, mock_exists):
        mock_open.return_value.__enter__ = lambda s: s
        mock_open.return_value.__exit__ = Mock()
        mock_open.return_value.read.return_value = RAW_ENVIRONMENTS_YAML
        mock_exists.return_value = True

        envs = environments()
        self.assertEqual(yaml.safe_load(RAW_ENVIRONMENTS_YAML), envs)

    @patch('os.path.isfile')
    def test_environments_noenv(self, mock_exists):
        mock_exists.return_value = False

        self.assertRaises(IOError, environments)

    @patch('amulet.helpers.environments')
    def test_default_environment(self, menvironments):
        menvironments.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)
        default = default_environment()
        self.assertEqual('gojuju', default)

    @patch('amulet.helpers.environments')
    def test_default_environment_no_default(self, menvironments):
        environments_yaml = """
        environments:
          gojuju1:
            type: cloud
            access-key: xxx
            secret-key: yyy
            control-bucket: gojuju-xyxyz
            admin-secret: zyxyx
            default-series: world"""
        menvironments.return_value = yaml.safe_load(environments_yaml)
        self.assertEqual('gojuju1', default_environment())

    @patch('amulet.helpers.environments')
    def test_default_environment_no_default_multi_fail(self, menvironments):
        envs = yaml.safe_load(RAW_ENVIRONMENTS_YAML)
        del envs['default']
        menvironments.return_value = envs

        self.assertRaises(ValueError, default_environment)


class JujuTest(unittest.TestCase):
    def test_juju(self):
        self.assertEqual(str(JujuVersion()), juju(['version']).split('-')[0])

    @patch('amulet.helpers.subprocess.Popen')
    def test_juju_oserror(self, mp):
        mp.side_effect = [OSError(1, 'Command Failed')]
        self.assertRaisesRegexp(OSError, 'Command Failed', juju, ['version'])


class RaiseStatusTest(unittest.TestCase):
    @patch('amulet.helpers.sys.exit')
    def test_raise_status(self, me):
        raise_status(0)
        me.assert_called_with(0)

    @patch('amulet.helpers.sys.exit')
    @patch('builtins.print' if sys.version_info > (3,) else '__builtin__.print')
    def test_raise_status_msg(self, mp, me):
        raise_status(100, 'Hello World')
        mp.assert_called_with('Hello World')
        me.assert_called_with(100)



import atexit
import os
import shlex
import shutil
import subprocess
import tempfile
import yaml

from .helpers import reify
from .helpers import run_bzr
from charmworldlib.charm import Charm
from path import path
from path import tempdir


class CharmCache(dict):
    def __init__(self, test_charm):
        super(CharmCache, self).__init__()
        self.test_charm = test_charm

    @staticmethod
    def get_charm(charm_path, branch=None, series='precise'):
        if charm_path.startswith('lp:'):
            return LaunchpadCharm(charm_path)
        elif branch and branch.startswith('lp:'):
            return LaunchpadCharm(branch)

        if charm_path.startswith('local:'):
            return LocalCharm(
                os.path.join(
                    os.environ.get('JUJU_REPOSITORY', ''),
                    charm_path[len('local:'):]),
                series
            )

        if branch and branch.endswith('.git'):
            return GitCharm(branch, name=charm_path)

        if os.path.exists(os.path.expanduser(charm_path)):
            return LocalCharm(charm_path, series)

        return Charm(with_series(charm_path, series))

    def __getitem__(self, service):
        return self.fetch(service)

    def fetch(self, service, charm=None, branch=None, series='precise'):
        charm_ = charm
        charm = super(CharmCache, self).get(service, None)
        if charm is not None:
            return charm

        charm = charm_ or service
        charm = os.getcwd() if charm == self.test_charm else charm
        self[service] = self.get_charm(charm,
                                       branch=branch,
                                       series=series)
        return self.get(service)


def with_series(charm_path, series):
    if '/' not in charm_path:
        return '{}/{}'.format(series, charm_path)
    return charm_path


class LocalCharm(object):
    def __init__(self, path, series):
        path = os.path.abspath(os.path.expanduser(path))

        if not os.path.exists(os.path.join(path, 'metadata.yaml')):
            raise Exception('Charm not found')

        if os.path.basename(os.path.dirname(path)) != series:
            path = self._make_temp_copy(path, series)

        self.url = path
        self.subordinate = False
        self.relations = {}
        self.provides = {}
        self.requires = {}
        self.code_source = self.source = None
        self._raw = self._load(os.path.join(path, 'metadata.yaml'))
        self._parse(self._raw)

    def _make_temp_copy(self, path, series):
        d = tempfile.mkdtemp(prefix='charm')
        atexit.register(shutil.rmtree, d)

        series_dir = os.path.join(d, series)
        os.mkdir(series_dir)
        temp_charm_dir = os.path.join(series_dir, os.path.basename(path))

        def ignore(src, names):
            return ['.git', '.bzr']

        shutil.copytree(path, temp_charm_dir, symlinks=True, ignore=ignore)
        return temp_charm_dir

    def _parse(self, metadata):
        rel_keys = ['provides', 'requires']
        for key, val in metadata.items():
            if key in rel_keys:
                self.relations[key] = val

            setattr(self, key, val)

    def _load(self, metadata_path):
        with open(metadata_path) as f:
            data = yaml.safe_load(f.read())

        return data

    def __str__(self):
        return yaml.dump(self._raw)

    def __repr__(self):
        return '<LocalCharm %s>' % self.url


class VCSCharm(object):

    def _parse(self, metadata):
        rel_keys = ['provides', 'requires']
        for key, val in metadata.items():
            if key in rel_keys:
                self.relations[key] = val

            setattr(self, key, val)

    def __str__(self):
        return yaml.dump(self._raw)


class GitCharm(VCSCharm):
    call = staticmethod(subprocess.check_call)

    def __init__(self, fork, name=None):
        self.name = name
        self.url = None
        self.subordinate = False
        self.code_source = self.source = {'location': fork, 'type': 'git'}
        self.relations = {}
        self.provides = {}
        self.requires = {}
        self.branch = self.fork = fork
        self._parse(self._raw)

    @reify
    def _raw(self):
        with tempdir() as td:
            cmd = "git clone -n --depth=1 {} {}"\
                .format(self.fork, self.name)

            with path(td):
                self.call(shlex.split(cmd))

            cmd = "git checkout HEAD metadata.yaml"
            with td / self.name:
                self.call(shlex.split(cmd))

            md = td / self.name / 'metadata.yaml'
            txt = md.text()
        return yaml.safe_load(txt)

    def __repr__(self):
        return '<GitCharm %s>' % self.code_source['location']


class LaunchpadCharm(VCSCharm):
    def __init__(self, branch):
        self.url = None
        self.subordinate = False
        self.code_source = self.source = {'location': branch, 'type': 'bzr'}
        self.relations = {}
        self.provides = {}
        self.requires = {}
        self._branch = branch
        self._parse(self._raw)

    @reify
    def _raw(self):
        metadata_path = path(self._branch) / 'metadata.yaml'
        mdata = run_bzr(['cat', metadata_path], None)
        return yaml.safe_load(mdata)

    def __repr__(self):
        return '<LaunchpadCharm %s>' % self.code_source['location']



import base64
import contextlib
import json
import logging
import os
import shlex
import subprocess
import yaml

from path import path
from path import tempdir

from .helpers import default_environment, juju, timeout as unit_timesout
from .sentry import Talisman
from .charm import CharmCache

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_charm_name(dir_):
    """Given a directory, return the name of the charm in that dir.

    If metadata.yaml exists in the dir, grab the charm name from there.
    Otherwise, return the name of the dir.

    """
    try:
        with open(os.path.join(dir_, 'metadata.yaml')) as f:
            return yaml.load(f)['name']
    except:
        return os.path.basename(dir_)


class Deployment(object):
    log = logger

    def __init__(self, juju_env=None, series='precise',
                 juju_deployer='juju-deployer', **kw):
        self.services = {}
        self.relations = []
        self.interfaces = []
        self.subordinates = []
        self.series = series
        self.deployed = False
        self.juju_env = juju_env or default_environment()
        self.charm_name = get_charm_name(os.getcwd())

        self.sentry = None
        self.deployer = path(juju_deployer)
        self.deployer_dir = tempdir(prefix='amulet_deployment_')

        if 'JUJU_TEST_CHARM' in os.environ:
            self.charm_name = os.environ['JUJU_TEST_CHARM']

        self.charm_cache = CharmCache(self.charm_name)

    @classmethod
    def from_bundle(cls, bundle_file, deployment_name=None):
        deployment = cls()
        bundle_file = path(bundle_file)
        deployment.load_bundle_file(bundle_file, deployment_name)
        return deployment

    def load_bundle_file(self, bundle_file, deployment_name=None):
        with open(bundle_file, 'r') as stream:
            contents = yaml.safe_load(stream)
        return self.load(contents, deployment_name)

    def load(self, deploy_cfg, deployment_name=None):
        schema = deploy_cfg.get(deployment_name, None) \
            or next(iter(deploy_cfg.values()))
        self.series = schema['series']
        self.relations = schema.get('relations', [])
        for service, service_config in schema['services'].items():
            constraints = service_config.get('constraints')
            if constraints:
                constraints = dict(
                    constraint.split('=')
                    for constraint in constraints.split()
                )

            self.add(
                service,
                charm=service_config.get('charm'),
                units=service_config.get('num_units', 1),
                branch=service_config.get('branch', None),
                constraints=constraints,
                placement=service_config.get('to', None),
                series=self.series
            )

            if service_config.get('options'):
                self.configure(service, service_config['options'])

            if service_config.get('expose'):
                self.expose(service)

    def add(self, service_name,
            charm=None,
            units=1,
            constraints=None,
            branch=None,
            placement=None,
            series=None):

        if self.deployed:
            raise NotImplementedError('Environment already setup')

        if service_name in self.services:
            raise ValueError('Service is already set to be deployed')

        service = self.services[service_name] = {}

        charm = self.charm_cache.fetch(
            service_name, charm, branch=branch, series=series or self.series)

        if charm.subordinate:
            for rtype in ['provides', 'requires']:
                try:
                    rels = getattr(charm, rtype)
                    for relation in rels:
                        rel = rels[relation]
                        if 'scope' in rel and rel['scope'] == 'container':
                            self.subordinates.append('%s:%s' %
                                                     (service_name, relation))
                except:  # @@ why is this diaper here?
                    pass

        source = charm.url and {'charm': charm.url} \
            or {'branch': charm.code_source['location']}

        service.update(source)

        service['num_units'] = units
        if placement is not None:
            service['to'] = placement

        if 'JUJU_TEST_CONSTRAINTS' in os.environ:
            env_constraints = {}
            for c in os.environ['JUJU_TEST_CONSTRAINTS'].split():
                try:
                    k, v = c.split('=')
                    env_constraints[k] = v
                except:
                    raise ValueError(
                        'Invalid constraint in JUJU_TEST_CONSTRAINTS: '
                        '%s' % (c))
            if constraints is not None:
                env_constraints.update(constraints)
            constraints = env_constraints

        if constraints is not None:
            if not isinstance(constraints, dict):
                raise ValueError('Constraints must be specified as a dict')

            r = ["%s=%s" % (k, v) for k, v in constraints.items()]
            service['constraints'] = " ".join(r)

        return service

    def add_unit(self, service, units=1, target=None):
        if not isinstance(units, int) or units < 1:
            raise ValueError('Only positive integers can be used for units')
        if target is not None and units != 1:
            raise ValueError(
                "Can't deploy more than one unit when specifying a target.")
        if service not in self.services:
            raise ValueError('Service needs to be added before you can scale')

        self.services[service]['num_units'] = \
            self.services[service].get('num_units', 1) + units

        if self.deployed:
            args = ['add-unit', service, '-n', str(units)]
            if target is not None:
                args.extend(["--to", target])
            juju(args)
            self.sentry = Talisman(self.services)

    def remove_unit(self, *units):
        if not self.deployed:
            raise NotImplementedError('Environment not setup yet')
        if not units:
            raise ValueError('No units provided')
        for unit in units:
            if '/' not in unit:
                raise ValueError('%s is not a unit' % unit)
            service = unit.split('/')[0]
            if service not in self.services:
                raise ValueError('%s is not a deployed service' % service)

        juju(['remove-unit'] + list(units))

        for unit in units:
            if self.sentry and unit in self.sentry.unit:
                del self.sentry.unit[unit]
            service = unit.split('/')[0]
            self.services[service]['num_units'] = \
                min(0, self.services[service].get('num_units', 1) - 1)
    destroy_unit = remove_unit

    def remove_service(self, *services):
        if not services:
            raise ValueError('No services provided')
        for service in services:
            if service not in self.services:
                raise ValueError('%s is not a deployed service' % service)

        if self.deployed:
            juju(['remove-service'] + list(services))

        for service in services:
            self._remove_service_sentries(service)
            self._remove_service_relations(service)
            del self.services[service]
    destroy_service = remove_service

    def remove(self, *units_or_services):
        if not units_or_services:
            raise ValueError('No units or services provided')

        units_or_services = set(units_or_services)
        units = {s for s in units_or_services if '/' in s}
        services = units_or_services - units
        units = {u for u in units if u.split('/')[0] not in services}

        if units:
            self.remove_unit(*units)

        if services:
            self.remove_service(*services)
    destroy = remove

    def _remove_service_sentries(self, service):
        if not self.sentry:
            return

        for unit in list(self.sentry.unit):
            if unit.split('/')[0] == service:
                del self.sentry.unit[unit]

    def _remove_service_relations(self, service):
        for relation in self.relations[:]:
            for rel_service in relation:
                if rel_service.split(':')[0] == service:
                    self.relations.remove(relation)
                    break

    def relate(self, *args):
        if len(args) < 2:
            raise LookupError('Need at least two services:relation')

        for srv_rel in args:
            if not ':' in srv_rel:
                raise ValueError('All relations must be explicit, ' +
                                 'service:relation')

            srv, rel = srv_rel.split(':')
            if srv not in self.services:
                raise ValueError('Can not relate, service not deployed yet')

            c = self.charm_cache[srv]
            if rel not in list(c.provides.keys()) + list(c.requires.keys()) \
                    + ['juju-info']:
                raise ValueError('%s does not exist for %s' % (rel, srv))

        args = list(args)
        first = args.pop(0)
        for srv in args:
            self._relate(first, srv)

    def _relate(self, a, b):
        if [a, b] not in self.relations and [b, a] not in self.relations:
            self.relations.append([a, b])
            if self.deployed:
                juju(['add-relation'] + [a, b])

    def unrelate(self, *args):
        if len(args) != 2:
            raise LookupError('Need exactly two service:relations')

        for srv_rel in args:
            if not ':' in srv_rel:
                raise ValueError('All relations must be explicit, ' +
                                 'service:relation')
        relation = list(args)
        for rel in relation, reversed(relation):
            if rel in self.relations:
                relation = rel
                break
        else:
            raise ValueError('Relation does not exist')

        self.relations.remove(relation)
        if self.deployed:
            juju(['remove-relation'] + relation)

    def schema(self):
        return self.deployer_map(self.services, self.relations)

    def configure(self, service, options):
        for k, v in options.items():
            include_token = 'include-base64://'
            if type(v) is str and v.startswith(include_token):
                v = v.replace(include_token, '')
                with open(os.path.join(os.getcwd(), 'tests', v)) as f:
                    v = base64.b64encode(f.read())
                service['options'][k] = v

        if self.deployed:
            opts = ['set', service]
            for k, v in options.items():
                opts.append("%s=%s" % (k, v))
            return juju(opts)

        if service not in self.services:
            raise ValueError('Service has not yet been described')

        if not 'options' in self.services[service]:
            self.services[service]['options'] = options
        else:
            self.services[service]['options'].update(options)

    def expose(self, service):
        if self.deployed:
            return juju(['expose', service])

        if service not in self.services:
            raise ValueError('%s has not yet been described' % service)
        self.services[service]['expose'] = True

    @contextlib.contextmanager
    def deploy_w_timeout_and_dir(self, timeout, deploy_dir):
        """
        :param timeout: Amount of time to wait for deployment to complete.
        :param deploy_dir: working directory for deployment command to run

        Sets timeout and working directory for wrapped block. If successful,
        sets instance.deployed.
        """
        with self.deployer_dir, unit_timesout(timeout):
            yield
        self.deployed = True

    def action_defined(self, service):
        """
        :param service: Name of service to get list of defined actions for.

        Returns the list of actions defined for the service.
        """
        if service not in self.services:
            raise ValueError(
                'Service needs to be added before listing actions.')
        raw = juju(['action', 'defined', service, '--format', 'json'])
        return json.loads(raw)

    def action_do(self, unit, action, action_args={}):
        """
        :param unit: Unit to run action on.
        :param action: Action to run.
        :param action_args: Action parameters.

        Runs specified action on specified unit and returns the uuid to fetch
        results by.

        """
        if '/' not in unit:
            raise ValueError('%s is not a unit' % unit)
        cmd = ['action', 'do', unit, action, '--format', 'json']
        for key, value in action_args.items():
            cmd += ["%s=%s" % (str(key), str(value))]
        result = juju(cmd)
        action_result = json.loads(result)
        results_id = action_result["Action queued with id"]
        return results_id

    def action_fetch(self, action_id, timeout=600):
        """
        :param action_id: Id of the action to fetch results for.
        """
        cmd = ['action', 'fetch', action_id, '--format', 'json']
        if timeout is not None:
            cmd += ["--wait", str(timeout)]
        raw = juju(cmd)
        result = json.loads(raw)
        status = result['status']
        if status == 'completed':
            if 'results' in result:
                return result['results']
        return {}

    def setup(self, timeout=600, cleanup=True):
        """Deploy the workload.

        :param timeout: Amount of time to wait for deployment to complete.
        :param cleanup: Set to False to leave the generated deployer file
            on disk. Useful for debugging.

        """
        if not self.deployer:
            raise NameError('Path to juju-deployer is not defined.')

        with tempdir(prefix='amulet-juju-deployer-') as tmpdir:
            schema_json = json.dumps(self.schema(), indent=2)
            self.log.debug("Deployer schema\n%s", schema_json)

            schema_file = tmpdir / 'deployer-schema.json'
            schema_file.write_text(schema_json)

            cmd = "{deployer} -W {debug} -c {schema} -e {env} -t {timeout} {env}"
            cmd_args = dict(
                deployer=self.deployer.expanduser(),
                debug=(
                    '-d'
                    if self.log.getEffectiveLevel() == logging.DEBUG
                    else ''),
                schema=schema_file,
                env=self.juju_env,
                timeout=str(timeout + 100),
            )
            cmd = cmd.format(**cmd_args)
            self.log.debug(cmd)

            with self.deploy_w_timeout_and_dir(timeout, self.deployer_dir):
                subprocess.check_call(shlex.split(cmd))

        self.sentry = Talisman(self.services)
        if cleanup is False:
            tmpdir.makedirs()
            (tmpdir / 'deployer-schema.json').write_text(schema_json)

    def deployer_map(self, services, relations):
        deployer_map = {
            self.juju_env: {
                'series': self.series,
                'services': self.services,
                'relations': self.build_relations()
            }
        }

        return deployer_map

    def build_relations(self):
        relations = []
        for rel in self.relations:
            relations.append(rel)

        return relations



#!/usr/bin/env python

import json
import os

PROC_DIR = '/proc'
JUJU_DIR = '/var/lib/juju/agents/'

d = {}
for pid in [p for p in os.listdir(PROC_DIR) if p.isdigit()]:
    try:
        cmd = open(os.path.join(PROC_DIR, pid, 'cmdline')).read()
    except:
        continue
    if JUJU_DIR in cmd:
        d['hook'] = os.path.basename(cmd)
        break

print(json.dumps(d))



import sys
import yaml

from .helpers import (
    TimeoutError,
    default_environment,
    timeout,
    juju,
    JujuVersion,
)

SUCCESS_STATES = ['started']


def wait(*args, **kwargs):
    """Wait until all criteria is met for a given juju environment

    When run without parameters the following defaults are used:
      - juju_env = os.environ['JUJU_ENV']
      - timeout = 300 (5m)
      - Wait will pause execution and wait for ALL units to move to a non-error
      state.

    You can specify which service or units you want to wait for by passing them
    as arguments:

      amulet.wait('service_a', 'service_a/1')

    """
    import os

    if not 'juju_env' in kwargs:
        kwargs['juju_env'] = \
            os.environ.get('JUJU_ENV') or default_environment()

    if not 'timeout' in kwargs:
        kwargs['timeout'] = 300

    ready = False
    with timeout(kwargs['timeout']):
        while not ready:
            try:
                raise_for_state(*args, juju_env=kwargs['juju_env'])
            except TimeoutError:
                raise
            except:
                ready = False
            else:
                ready = True

    return True


def raise_for_state(*args, **kwargs):
    status = state(*args, **kwargs)

    for service in status:
        for unit in status[service]:
            if not status[service][unit] in SUCCESS_STATES:
                raise StateError("%s: %s" % (unit, status[service][unit]))


class StateError(Exception):
    def __init__(self, value="A unit is in a pending or error state"):
        self.value = value


# Move these to another module?
def _get_gojuju_status(environment=None):
    return _get_pyjuju_status(environment)


# Move these to another module?
def _get_pyjuju_status(environment=None):
    cmd = ['status', '--format', 'yaml']
    if environment:
        cmd.extend(['-e', environment])

    try:
        status_yml = juju(cmd)
    except TimeoutError:
        raise
    except:
        raise Exception('Unable to query status for %s' % environment)

    return yaml.safe_load(status_yml)


def get_state(data):
    states = ['life', 'relations-error', 'agent-state']
    for state_key in states:
        if state_key in data:
            return str(data[state_key])


def status(juju_env=None):
    version = JujuVersion()

    if not juju_env:
        raise KeyError('No juju_env set')

    try:
        if version.major == 0:
            juju_status = _get_pyjuju_status(juju_env)
        else:
            juju_status = _get_gojuju_status(juju_env)
    except TimeoutError:
        raise
    except:
        raise

    return juju_status


def state(*args, **kwargs):
    output = {}

    if not 'juju_env' in kwargs:
        raise KeyError('No juju_env set')

    juju_env = kwargs['juju_env']

    try:
        juju_status = status(juju_env)
    except TimeoutError:
        raise
    except:
        return output

    if not args:
        args = [service for service in juju_status['services']]

    for arg in args:
        service, unit = arg.split('/') if '/' in arg else [arg, None]
        if not service in juju_status['services']:
            raise ValueError('%s is not in the deployment yet' % arg)

        if not service in output:
            output[service] = {}

        if not 'units' in juju_status['services'][service]:
            # Probably a subordinate
            if 'subordinate-to' in juju_status['services'][service]:
                del output[service]
            continue

        # Use potential recurive + mergedicts?
        # http://stackoverflow.com/a/7205672/196832
        units = juju_status['services'][service]['units']
        if unit:
            output[service][unit] = get_state(units['/'.join([service, unit])])
        else:
            for unit_name in units:
                unit = unit_name.split('/')[1]
                s = get_state(units['/'.join([service, unit])])
                output[service][unit] = s

    return output


def setup_parser(parent):
    def wait_cmd(args):
        try:
            wait(*args.services, **vars(args))
        except TimeoutError:
            sys.stderr.write('Timeout criteria was met\n')
            sys.exit(124)
        except:
            sys.stderr.write('Unexpected error occurred\n')
            raise

        sys.exit(0)

    parser = parent.add_parser('wait', help="Wait until criteria is met")
    parser.add_argument('-e', '--environment', dest='juju_env',
                        help="Juju environment")
    parser.add_argument('-t', '--timeout', help="Timeout in seconds", type=int,
                        default=300)
    parser.add_argument('services', nargs='*',
                        help="What services or units to wait on")
    parser.set_defaults(func=wait_cmd)



#!/usr/bin/env python

import json
import os
import sys

s = os.stat(sys.argv[1])
d = {
    'mtime': s.st_mtime,
    'size': s.st_size,
    'uid': s.st_uid,
    'gid': s.st_gid,
    'mode': oct(s.st_mode)
}
print(json.dumps(d))



#!/usr/bin/env python

import json
import os
import sys

path = sys.argv[1]

contents = {'files': [], 'directories': []}
for fd in os.listdir(path):
    if os.path.isfile('{}/{}'.format(path, fd)):
        contents['files'].append(fd)
    else:
        contents['directories'].append(fd)
print(json.dumps(contents))



"""Unit test for amulet.wait"""

import os
import unittest
import json
import shutil
import tempfile
import yaml

from amulet import Deployment
from amulet.deployer import CharmCache
from amulet.deployer import get_charm_name
from amulet.sentry import UnitSentry
from mock import patch, MagicMock, call
from collections import OrderedDict

RAW_ENVIRONMENTS_YAML = '''
default: gojuju
environments:
  gojuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: gojuju-xyxyz
    admin-secret: zyxyx
    default-series: world
  pyjuju:
    type: cloud
    access-key: xxx
    secret-key: yyy
    control-bucket: juju-xyxyz
    admin-secret: zyxyxy
    default-series: world
    juju-origin: ppa
    ssl-hostname-verification: true'''


class DeployerTests(unittest.TestCase):
    def test_init(self):
        d = Deployment(juju_env='gojuju')
        self.assertEqual('precise', d.series)
        self.assertEqual('gojuju', d.juju_env)
        self.assertEqual({}, d.services)
        self.assertEqual([], d.relations)

    def test_load(self):
        d = Deployment(juju_env='gojuju')
        schema = '{"mybundle": {"series": "raring", "services": {"wordpress": \
                  {"branch": "lp:~charmers/charms/precise/wordpress/trunk", \
                  "expose": true}, "mysql": {"options": {"tuning": \
                  "fastest"}, "constraints": "mem=2G cpu-cores=2", \
                  "branch": "lp:~charmers/charms/precise/mysql/trunk"}}, \
                  "relations": [["mysql:db", "wordpress:db"]]}}'
        dmap = json.loads(schema)
        with patch.object(d, 'add') as add:
            with patch.object(d, 'configure') as configure:
                with patch.object(d, 'expose') as expose:
                    d.load(dmap)
        self.assertEqual(d.juju_env, 'gojuju')
        self.assertEqual(dmap['mybundle']['relations'], d.relations)
        self.assertEqual(dmap['mybundle']['series'], d.series)
        add.assert_has_calls([
            call('wordpress',
                 placement=None,
                 series='raring',
                 units=1,
                 branch='lp:~charmers/charms/precise/wordpress/trunk',
                 constraints=None,
                 charm=None),
            call('mysql',
                 placement=None,
                 series='raring',
                 units=1,
                 branch='lp:~charmers/charms/precise/mysql/trunk',
                 constraints={'mem': '2G', 'cpu-cores': '2'},
                 charm=None)],
            any_order=True
        )
        configure.assert_has_calls([
            call('mysql', {'tuning': 'fastest'}),
        ])
        expose.assert_has_calls([
            call('wordpress'),
        ])

    @patch('amulet.charm.CharmCache.get_charm')
    def test_add(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location':
                             'lp:~charmers/charms/precise/charm/trunk'}
        charm.url = 'cs:precise/charm'

        d = Deployment(juju_env='gojuju')
        d.add('charm')
        self.assertEqual({'charm': {'charm': 'cs:precise/charm',
                                    'num_units': 1}}, d.services)

    @patch('amulet.charm.CharmCache.get_charm')
    def test_add_branch(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:~foo/charms/precise/baz/trunk'}
        charm.url = None

        d = Deployment(juju_env='gojuju')
        d.add('bar', 'cs:~foo/baz')
        self.assertEqual({'bar': {'branch': 'lp:~foo/charms/precise/baz/trunk',
                                  'num_units': 1}}, d.services)

    @patch('amulet.charm.CharmCache.get_charm')
    def test_add_units(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None

        d = Deployment(juju_env='gojuju')
        d.add('charm', units=2)
        self.assertEqual({'charm': {'branch': 'lp:charms/charm',
                                    'num_units': 2}}, d.services)

    @patch('amulet.charm.CharmCache.get_charm')
    def test_add_constraints(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None

        d = Deployment(juju_env='gojuju')
        d.add('charm', units=2, constraints=OrderedDict([
            ("cpu-power", 0),
            ("cpu-cores", 4),
            ("mem", "512M")
        ]))

        self.assertEqual({'charm': {'branch': 'lp:charms/charm',
                                    'constraints':
                                    'cpu-power=0 cpu-cores=4 mem=512M',
                                    'num_units': 2}}, d.services)

    def _make_mock_status(self, d):
        def _mock_status(juju_env):
            status = dict(services={})
            total_units = 1
            for service in d.services:
                status['services'][service] = dict(units={})
                for unit in range(d.services[service].get('num_units', 1)):
                    total_units += 1
                    status['services'][service]['units'][
                        '{}/{}'.format(service, unit)] = {
                        'agent-state': 'started',
                        'public-address': '10.0.3.{}'.format(total_units)}
            status['services']['relation-sentry'] = {
                'units': {
                    'relation-sentry/0': {
                        'agent-state': 'started',
                        'public-address': '10.0.3.1'}}}
            return status
        return _mock_status

    @patch.object(UnitSentry, 'upload_scripts')
    @patch('amulet.helpers.environments')
    @patch('amulet.sentry.waiter.status')
    @patch('amulet.deployer.subprocess')
    @patch('amulet.charm.CharmCache.get_charm')
    def test_add_unit(self, mcharm, subprocess, waiter_status, environments,
                      upload_scripts):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None
        charm.series = 'precise'

        environments.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)

        d = Deployment(juju_env='gojuju')

        waiter_status.side_effect = self._make_mock_status(d)
        d.add('charm', units=1)
        d.setup()
        with patch('amulet.deployer.juju') as j:
            d.add_unit('charm')
            j.assert_called_with(['add-unit', 'charm', '-n', '1'])
        self.assertTrue('charm/1' in d.sentry.unit)
        self.assertEqual(2, d.services['charm']['num_units'])

    @patch.object(UnitSentry, 'upload_scripts')
    @patch('amulet.helpers.environments')
    @patch('amulet.sentry.waiter.status')
    @patch('amulet.deployer.subprocess')
    @patch('amulet.charm.CharmCache.get_charm')
    def test_add_unit_target(self, mcharm, subprocess, waiter_status,
                             environments, upload_scripts):
        """
        If target is given to Deployment.add_unit(), the 'juju' call
        will includde --to to make the unit be added to the given
        target.
        """
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None
        charm.series = 'precise'

        environments.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)

        d = Deployment(juju_env='gojuju')

        waiter_status.side_effect = self._make_mock_status(d)
        d.add('charm', units=1)
        d.setup()
        with patch('amulet.deployer.juju') as j:
            d.add_unit('charm', target='lxc:0')
            j.assert_called_with(
                ['add-unit', 'charm', '-n', '1', '--to', 'lxc:0'])

    @patch.object(UnitSentry, 'upload_scripts')
    @patch('amulet.helpers.environments')
    @patch('amulet.sentry.waiter.status')
    @patch('amulet.deployer.subprocess')
    @patch('amulet.charm.CharmCache.get_charm')
    def test_add_unit_error(self, mcharm, subprocess, waiter_status,
                            environments, upload_scripts):
        def mock_unit_error(f, service, unit_name):
            def _mock_unit_error(juju_env):
                status = f(juju_env)
                unit = status['services'][service]['units'].get(unit_name)
                if not unit:
                    return status
                unit['agent-state'] = 'error'
                unit['agent-state-info'] = 'hook failed: install'
                return status
            return _mock_unit_error

        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location': 'lp:charms/charm'}
        charm.url = None
        charm.series = 'precise'

        environments.return_value = yaml.safe_load(RAW_ENVIRONMENTS_YAML)

        d = Deployment(juju_env='gojuju')
        waiter_status.side_effect = mock_unit_error(
            self._make_mock_status(d), 'charm', 'charm/1')
        d.add('charm', units=1)
        d.setup()
        with patch('amulet.deployer.juju'):
            self.assertRaisesRegexp(
                Exception, 'Error on unit charm/1: hook failed: install',
                d.add_unit, 'charm')

    @patch('amulet.charm.CharmCache.get_charm')
    def test_remove_unit(self, get_charm):
        d = Deployment(juju_env='gojuju')
        d.add('charm')
        patcher = patch.object(d, 'sentry', MagicMock(unit={'charm/0': 1}))
        self.addCleanup(patcher.stop)
        sentry = patcher.start()
        d.deployed = True
        with patch('amulet.deployer.juju') as juju:
            d.remove_unit('charm/0')
            juju.assert_called_once_with(['remove-unit', 'charm/0'])
        self.assertFalse('charm/0' in sentry.unit)
        self.assertEqual(0, d.services['charm']['num_units'])

    @patch('amulet.charm.CharmCache.get_charm')
    def test_remove_service(self, get_charm):
        d = Deployment(juju_env='gojuju')
        d.add('charm')
        patcher = patch.object(d, 'sentry', MagicMock(unit={'charm/0': 1}))
        self.addCleanup(patcher.stop)
        sentry = patcher.start()
        d.deployed = True
        d.relations = [('charm:rel', 'another:rel')]
        with patch('amulet.deployer.juju') as juju:
            d.remove_service('charm')
            juju.assert_called_once_with(['remove-service', 'charm'])
        self.assertFalse('charm/0' in sentry.unit)
        self.assertFalse('charm' in d.services)
        self.assertEqual(0, len(d.relations))

    @patch('amulet.charm.CharmCache.get_charm')
    def test_remove(self, get_charm):
        d = Deployment(juju_env='gojuju')
        d.add('charm1')
        d.add('charm2')
        p1 = patch.object(d, 'remove_unit')
        p2 = patch.object(d, 'remove_service')
        self.addCleanup(p1.stop)
        self.addCleanup(p2.stop)
        remove_unit = p1.start()
        remove_service = p2.start()
        with patch('amulet.deployer.juju'):
            d.remove('charm1/0', 'charm2')
            remove_unit.assert_called_once_with('charm1/0')
            remove_service.assert_called_once_with('charm2')

    @patch('amulet.charm.CharmCache.get_charm')
    def test_remove_aliases(self, get_charm):
        d = Deployment
        self.assertEqual(d.destroy_unit, d.remove_unit)
        self.assertEqual(d.destroy_service, d.remove_service)
        self.assertEqual(d.destroy, d.remove)

    @patch('amulet.charm.CharmCache.get_charm')
    def test_add_error(self, mcharm):
        d = Deployment(juju_env='gojuju')
        d.add('bar')
        self.assertRaises(ValueError, d.add, 'bar')

    @patch('amulet.charm.CharmCache.get_charm')
    def test_relate(self, mcharm):
        a = mcharm.return_value
        a.provides = {'f': {'interface': 'test'}}
        a.requires = {'b': {'interface': 'test'}}

        d = Deployment(juju_env='gojuju')
        d.add('bar')
        d.add('foo')
        d.relate('foo:f', 'bar:b')
        self.assertEqual(d.relations, [['foo:f', 'bar:b']])

    def test_relate_too_few(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(LookupError, d.relate, 'foo:f')

    def test_relate_no_relation(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.relate, 'foo', 'bar:a')

    @patch('amulet.charm.CharmCache.get_charm')
    def test_relate_relation_nonexist(self, mcharm):
        d = Deployment(juju_env='gojuju')
        d.add('bar')
        d.add('foo')

        self.assertRaises(ValueError, d.relate, 'foo:f', 'bar:b')

    def test_relate_not_deployed(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.relate, 'foo:f', 'bar:a')

    @patch('amulet.charm.CharmCache.get_charm')
    def test_configure(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location':
                             'lp:~charmers/charms/precise/wordpress/trunk'}
        charm.url = 'cs:precise/wordpress'

        d = Deployment(juju_env='gojuju')
        d.add('wordpress')
        d.configure('wordpress', {'tuning': 'optimized'})
        d.configure('wordpress', {'wp-content': 'f', 'port': 100})
        self.assertEqual({'wordpress':
                          {'charm': 'cs:precise/wordpress',
                           'num_units': 1,
                           'options': {'tuning': 'optimized',
                                       'wp-content': 'f',
                                       'port': 100}}}, d.services)

    def test_configure_not_deployed(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.configure, 'wordpress',
                          {'tuning': 'optimized'})

    @patch('amulet.charm.CharmCache.get_charm')
    def test_expose(self, mcharm):
        charm = mcharm.return_value
        charm.subordinate = False
        charm.code_source = {'location':
                             'lp:~charmers/charms/precise/wordpress/trunk'}
        charm.url = None

        d = Deployment(juju_env='gojuju')
        d.add('wordpress')
        d.expose('wordpress')
        self.assertEqual(
            {'wordpress':
                {'branch': 'lp:~charmers/charms/precise/wordpress/trunk',
                 'num_units': 1,
                 'expose': True}}, d.services)

    def test_expose_not_deployed(self):
        d = Deployment(juju_env='gojuju')
        self.assertRaises(ValueError, d.expose, 'wordpress')

    @patch('amulet.charm.CharmCache.get_charm')
    def test_schema(self, mcharm):
        wpmock = MagicMock()
        mysqlmock = MagicMock()
        wpmock.subordinate = False
        wpmock.code_source = {'location':
                              'lp:~charmers/charms/precise/wordpress/trunk'}
        wpmock.requires = {'db': {'interface': 'mysql'}}
        wpmock.url = None
        mysqlmock.subordinate = False
        mysqlmock.code_source = {'location':
                                 'lp:~charmers/charms/precise/mysql/trunk'}
        mysqlmock.provides = {'db': {'interface': 'mysql'}}
        mysqlmock.url = None

        mcharm.side_effect = [mysqlmock, wpmock]
        d = Deployment(juju_env='gojuju', sentries=False)
        d.add('mysql')
        d.configure('mysql', {'tuning': 'fastest'})
        d.add('wordpress')
        d.relate('mysql:db', 'wordpress:db')
        schema = {'gojuju': {
            'services': {
                'mysql': {
                    'branch': 'lp:~charmers/charms/precise/mysql/trunk',
                    'num_units': 1,
                    'options': {'tuning': 'fastest'},
                },
                'wordpress': {
                    'branch': 'lp:~charmers/charms/precise/wordpress/trunk',
                    'num_units': 1
                }
            },
            'series': 'precise',
            'relations': [['mysql:db', 'wordpress:db']],
        }}
        self.assertEqual(schema, d.schema())

    @patch.dict('os.environ', {'JUJU_TEST_CHARM': 'charmbook'})
    def test_juju_test_charm(self):
        d = Deployment(juju_env='gogo')
        self.assertEqual('charmbook', d.charm_name)

    def test_add_post_deploy(self):
        d = Deployment(juju_env='gogo')
        d.deployed = True
        self.assertRaises(NotImplementedError, d.add, 'mysql')

    def test_relate_before_service_deployed(self):
        d = Deployment(juju_env='gogo')
        d.deployed = True
        with self.assertRaises(ValueError) as e:
            d.relate('mysql:db', 'wordpress:db')
            self.assertEqual(
                'Can not relate, service not deployed yet', str(e))

    def test_unrelate_not_enough(self):
        d = Deployment(juju_env='gogo')
        self.assertRaises(LookupError, d.unrelate, 'mysql')

    def test_unrelate_too_many(self):
        d = Deployment(juju_env='gogo')
        self.assertRaises(LookupError, d.unrelate,
                          'mysql', 'wordpress', 'charm')

    def test_unrelate_bad_service_format(self):
        d = Deployment(juju_env='gogo')
        d.deployed = True
        with self.assertRaises(ValueError) as e:
            d.unrelate('mysql', 'wordpress')
            self.assertEqual(
                str(e), 'All relations must be explicit, service:relation')

    @patch('amulet.deployer.juju')
    def test_action_defined(self, mj):
        mj.return_value = '{"action": "description"}'
        d = Deployment(juju_env='gojuju')
        d.add('mysql')
        actions = d.action_defined('mysql')
        self.assertEquals(actions, {"action": "description"})
        mj.assert_has_calls([call(['action', 'defined', 'mysql', '--format', 'json'])])

    @patch('amulet.deployer.juju')
    def test_action_do(self, mj):
        mj.return_value = '{"Action queued with id": "some-action-id"}'
        d = Deployment(juju_env='gojuju')
        d.add('mysql')
        uuid = d.action_do('mysql/0', 'run')
        self.assertEquals(uuid, "some-action-id")
        mj.assert_has_calls([call(['action', 'do', 'mysql/0', 'run', '--format', 'json'])])

    @patch('amulet.deployer.juju')
    def test_action_fetch(self, mj):
        mj.side_effect = ['{"Action queued with id": "some-action-id"}',
                          '{"results":{"key":"value"},"status":"completed",'\
                          '"timing":{"completed":"2015-07-21 09:05:11 +0300 EEST",'\
                          '"enqueued":"2015-07-21 09:05:06 +0300 EEST",'\
                          '"started":"2015-07-21 09:05:09 +0300 EEST"}}']
        d = Deployment(juju_env='gojuju')
        d.add('mysql')
        uuid = d.action_do('mysql/0', 'run', {"action_param": "action_value"})
        self.assertEquals(uuid, "some-action-id")
        results = d.action_fetch(uuid)
        self.assertEquals(results, {'key': 'value'})
        mj.assert_has_calls([call(['action', 'do', 'mysql/0', 'run', '--format', 'json', 'action_param=action_value']),
                         call(['action', 'fetch', 'some-action-id', '--format', 'json', '--wait', '600'])])

    @patch('amulet.deployer.juju')
    def test_action_fetch_nowait_fail(self, mj):
        mj.side_effect = ['{"Action queued with id": "some-action-id"}',
                          '{"status":"running",'\
                          '"timing":{"enqueued":"2015-07-21 09:50:59 +0300 EEST",'\
                          '"started":"2015-07-21 09:51:04 +0300 EEST"}}']
        d = Deployment(juju_env='gojuju')
        d.add('mysql')
        uuid = d.action_do('mysql/0', 'run')
        self.assertEquals(uuid, "some-action-id")
        results = d.action_fetch(uuid, timeout=None)
        self.assertEquals(results, {})
        mj.assert_has_calls([call(['action', 'do', 'mysql/0', 'run', '--format', 'json']),
                             call(['action', 'fetch', 'some-action-id', '--format', 'json'])])

    @patch('amulet.deployer.juju')
    def test_action_fetch_wait(self, mj):
        mj.side_effect = ['{"Action queued with id": "some-action-id"}',
                          '{"results":{"key":"value"},"status":"completed",'\
                          '"timing":{"completed":"2015-07-21 09:05:11 +0300 EEST",'\
                          '"enqueued":"2015-07-21 09:05:06 +0300 EEST",'\
                          '"started":"2015-07-21 09:05:09 +0300 EEST"}}']
        d = Deployment(juju_env='gojuju')
        d.add('mysql')
        uuid = d.action_do('mysql/0', 'run')
        self.assertEquals(uuid, "some-action-id")
        results = d.action_fetch(uuid)
        self.assertEquals(results, {'key': 'value'})
        mj.assert_has_calls([call(['action', 'do', 'mysql/0', 'run', '--format', 'json']),
                             call(['action', 'fetch', 'some-action-id', '--format', 'json', '--wait', '600'])])
        
    @patch('amulet.deployer.juju')
    def test_unrelate(self, mj):
        d = Deployment(juju_env='gogo')
        d._relate('mysql:db', 'charm:db')
        d.deployed = True
        d.unrelate('mysql:db', 'charm:db')
        mj.assert_has_calls([
            call(['remove-relation',
                  'mysql:db', 'charm:db']),
            ])

    def test_unrelate_post_deploy(self):
        d = Deployment(juju_env='gogo')
        with self.assertRaises(ValueError) as e:
            d.unrelate('mysql:db', 'wordpress:db')
            self.assertEqual('Relation does not exist', str(e))

    @patch('amulet.deployer.juju')
    def test_remove_unit(self, mj):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        d.remove_unit('mysql/1')
        mj.assert_called_with(['remove-unit', 'mysql/1'])

    def test_remove_unit_env_not_setup(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        self.assertRaises(NotImplementedError, d.remove_unit, 'mysql/0')

    def test_remove_unit_no_args(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        self.assertRaises(ValueError, d.remove_unit)

    def test_remove_unit_not_a_unit(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        self.assertRaises(ValueError, d.remove_unit, 'mysql/1', 'lolk')

    def test_remove_unit_not_deployed(self):
        d = Deployment(juju_env='gogo')
        d.add('mysql', units=2)
        d.deployed = True
        self.assertRaises(ValueError, d.remove_unit, 'wordpress/1')






class GetCharmNameTest(unittest.TestCase):
    def setUp(self):
        self.dir_ = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.dir_)

    def test_name_from_yaml(self):
        """Charm name pulled from metadata.yaml if it exists"""
        with open(os.path.join(self.dir_, 'metadata.yaml'), 'w') as f:
            f.write('name: testcharm')

        self.assertEqual(get_charm_name(self.dir_), 'testcharm')

    def test_name_from_dir(self):
        """Charm name equals dir name if no metadata.yaml exists"""
        self.assertEqual(
            get_charm_name(self.dir_), os.path.basename(self.dir_))



import os
import tempfile
import unittest
import yaml

from mock import patch

from amulet.charm import CharmCache
from amulet.charm import LocalCharm
from amulet.charm import run_bzr


class RunBzrTest(unittest.TestCase):
    def test_run_bzr(self):
        out = run_bzr(["rocks"], ".")
        self.assertEquals(out, "It sure does!\n")

    @patch('subprocess.Popen')
    def test_run_bzr_traceback(self, mp):
        mp.side_effect = [Exception("AssertionError: always fails")]
        self.assertRaisesRegexp(Exception, "AssertionError: always fails",
                                run_bzr, ["assert-fail"], ".")

    @patch('subprocess.Popen')
    def test_run_bzr_oserror(self, mp):
        mp.side_effect = [OSError(1, "Command failed")]
        self.assertRaisesRegexp(OSError, "Command failed",
                                run_bzr, ["assert-fail"], ".")

    def test_run_bzr_missing(self):
        env = os.environ.copy()
        env["PATH"] = ""
        self.assertRaisesRegexp(OSError, "bzr not found, do you have Bazaar "
                                "installed?", run_bzr, ["version"], ".",
                                env=env)


class LocalCharmTest(unittest.TestCase):
    @patch('os.path.exists')
    @patch.object(LocalCharm, '_load')
    def test_parse(self, mlc_l, me):
        me.return_value = True
        mlc_l.return_value = {'name': 'test',
                              'author': 'ohkay',
                              'provides': {'rel': {'interface': 'test'}}}
        c = LocalCharm('/path/to/precise/test', 'precise')
        self.assertEqual('test', c.name)
        mlc_l.assert_called_with('/path/to/precise/test/metadata.yaml')

    def test_make_temp(self):
        charm_dir = tempfile.mkdtemp()
        metadata = {}
        with open(os.path.join(charm_dir, 'metadata.yaml'), 'w') as f:
            f.write(yaml.dump(metadata))

        c = LocalCharm(charm_dir, 'trusty')
        code_source = c.url
        self.assertTrue(code_source != charm_dir)
        self.assertTrue(code_source.startswith('/tmp/charm'))
        self.assertEqual(
            os.path.basename(os.path.dirname(code_source)),
            'trusty')
        self.assertTrue(os.path.exists(os.path.join(
            code_source, 'metadata.yaml')))


class LaunchpadCharmTest(unittest.TestCase):
    pass


class GetCharmTest(unittest.TestCase):
    def test_local(self):
        with patch('amulet.charm.LocalCharm') as LocalCharm:

            # Patch w/o JUJU_REPOSITORY
            with patch.dict('amulet.charm.os.environ', {
                    'JUJU_REPOSITORY': ''}):
                CharmCache.get_charm('local:precise/mycharm')
                LocalCharm.assert_called_once_with(
                    'precise/mycharm', 'precise')
                LocalCharm.reset_mock()

            # Patch w/JUJU_REPOSITORY
            with patch.dict('amulet.charm.os.environ', {
                    'JUJU_REPOSITORY': '~/charms'}):
                CharmCache.get_charm('local:precise/mycharm')
                LocalCharm.assert_called_once_with(
                    '~/charms/precise/mycharm', 'precise')


class CharmCacheTest(unittest.TestCase):
    def test_init(self):
        c = CharmCache('mytestcharm')
        self.assertEqual(c.test_charm, 'mytestcharm')

    def test_getitem_service(self):
        c = CharmCache('mytestcharm')
        with patch.object(c, 'get_charm') as get_charm:
            charm = c['myservice']
            self.assertEqual(charm, get_charm.return_value)
            get_charm.assert_called_once_with('myservice', branch=None,
                                              series='precise')
            get_charm.reset_mock()
            charm2 = c['myservice']
            self.assertEqual(charm, charm2)
            self.assertFalse(get_charm.called)

    def test_getitem_testcharm(self):
        c = CharmCache('mytestcharm')
        with patch.object(c, 'get_charm') as get_charm:
            charm = c['mytestcharm']
            self.assertEqual(charm, get_charm.return_value)
            get_charm.assert_called_once_with(os.getcwd(), branch=None,
                                              series='precise')

    def test_fetch_service(self):
        c = CharmCache('mytestcharm')
        with patch.object(c, 'get_charm') as get_charm:
            charm = c.fetch('myservice')
            self.assertEqual(charm, get_charm.return_value)
            get_charm.assert_called_once_with('myservice', branch=None,
                                              series='precise')

        get_charm.reset_mock()
        charm2 = c['myservice']
        self.assertEqual(charm, charm2)
        self.assertFalse(get_charm.called)

    def test_fetch_charm(self):
        c = CharmCache('mytestcharm')
        with patch.object(c, 'get_charm') as get_charm:
            charm = c.fetch('myservice', 'anothercharm')
            self.assertEqual(charm, get_charm.return_value)
            get_charm.assert_called_once_with(
                'anothercharm', branch=None, series='precise')

    def test_fetch_testcharm(self):
        c = CharmCache('mytestcharm')
        with patch.object(c, 'get_charm') as get_charm:
            charm = c.fetch('myservice', 'mytestcharm')
            self.assertEqual(charm, get_charm.return_value)
            get_charm.assert_called_once_with(os.getcwd(), branch=None,
                                              series='precise')



import amulet
import unittest


class TestDeployment(unittest.TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.deployment = amulet.Deployment(series='precise')

        cls.deployment.add('nagios')
        cls.deployment.add('haproxy')
        cls.deployment.add('rsyslog-forwarder')
        cls.deployment.relate('nagios:website', 'haproxy:reverseproxy')
        cls.deployment.relate('nagios:juju-info', 'rsyslog-forwarder:juju-info')

        try:
            cls.deployment.setup(timeout=900)
            cls.deployment.sentry.wait()
        except amulet.helpers.TimeoutError:
            amulet.raise_status(
                amulet.SKIP, msg="Environment wasn't stood up in time")
        except:
            raise

        cls.nagios = cls.deployment.sentry['nagios/0']
        cls.haproxy = cls.deployment.sentry['haproxy/0']
        cls.rsyslogfwd = cls.deployment.sentry['rsyslog-forwarder/0']
        cls.nagios.run(
            'mkdir -p /tmp/amulet-test/test-dir;'
            'echo contents > /tmp/amulet-test/test-file;'
        )
        cls.rsyslogfwd.run(
            'echo more-contents > /tmp/amulet-sub-test;'
        )


    def test_add_unit(self):
         self.deployment.add_unit('haproxy')
         haproxy = self.deployment.sentry['haproxy/1']
         self.assertEqual('1', haproxy.info['unit'])
         self.assertEqual('haproxy/1', haproxy.info['unit_name'])


    def test_info(self):
        self.assertTrue('public-address' in self.nagios.info)
        self.assertEqual('nagios', self.nagios.info['service'])
        self.assertEqual('0', self.nagios.info['unit'])
        self.assertEqual('nagios/0', self.nagios.info['unit_name'])

    def test_file_stat(self):
        path = '/tmp/amulet-test/test-file'
        stat = self.nagios.file_stat(path)
        self.assertTrue(stat.pop('mtime'))
        self.assertEqual(
            stat, {
                'size': 9,
                'uid': 0,
                'gid': 0,
                'mode': '0100644',
            },
        )

    def test_file_contents(self):
        path = '/tmp/amulet-test/test-file'
        self.assertEqual(
            self.nagios.file_contents(path),
            'contents\n',
        )

    def test_subordinate_file_contents(self):
        path = '/tmp/amulet-sub-test'
        self.assertEqual(
            self.rsyslogfwd.file_contents(path),
            'more-contents\n',
        )

    def test_directory_stat(self):
        path = '/tmp/amulet-test'
        stat = self.nagios.directory_stat(path)
        self.assertTrue(stat.pop('mtime'))

        """
        The block size is dependent on the file system used.
        btrfs defaults to 16k, ext2/3/4 defaults to 4k, etc.
        In this case, trust the stat size returned from the unit.
        """
        self.assertEqual(
            stat, {
                'size': stat['size'],
                'uid': 0,
                'gid': 0,
                'mode': '040755',
            },
        )

    def test_directory_listing(self):
        path = '/tmp/amulet-test'
        self.assertEqual(
            self.nagios.directory_listing(path), {
                'files': ['test-file'],
                'directories': ['test-dir'],
            },
        )

    def test_relation(self):
        nagios_info = self.nagios.relation(
            'website', 'haproxy:reverseproxy')
        for key in ['hostname', 'port', 'private-address']:
            self.assertTrue(key in nagios_info)

        haproxy_info = self.haproxy.relation(
            'reverseproxy', 'nagios:website')
        self.assertEqual(list(haproxy_info.keys()), ['private-address'])

    def test_run(self):
        self.assertEqual(
            self.nagios.run('echo hello'),
            ('hello', 0),
        )


if __name__ == '__main__':
    unittest.main()



from .waiter import wait

from .deployer import Deployment
from .helpers import (
    FAIL,
    PASS,
    SKIP,
    TimeoutError,
    default_environment,
    fail_if_timeout,
    raise_status,
    timeout,
)



import unittest


class TestGitCham(unittest.TestCase):

    def makeone(self, fork="http://github.com/juju-solutions/kraken.git",
                name="gitty"):
        from amulet.charm import GitCharm
        return GitCharm(fork, name)



import glob
import json
import os
import subprocess
import time

import pkg_resources

from . import waiter
from . import helpers


class SentryError(Exception):
    pass


class Sentry(object):
    def __init__(self, address, port=9001):
        self.config = {}
        self.config['address'] = 'https://%s:%s' % (address, port)

    def file(self, filename):
        return self.file_stat(filename)

    def file_stat(self, filename):
        raise NotImplemented()

    def file_contents(self, filename):
        raise NotImplemented()

    def directory(self, *args):
        return self.directory_stat(*args)

    def directory_stat(self, *args):
        raise NotImplemented()

    def directory_contents(self, *args):
        return self.directory_listing(*args)

    def directory_listing(self, *args):
        raise NotImplemented()

    def juju_agent(self, timeout):
        raise NotImplemented()


class UnitSentry(Sentry):
    @classmethod
    def fromunit(cls, unit):
        pass

    @classmethod
    def fromunitdata(cls, unit, unit_data):
        address = unit_data['public-address']
        unitsentry = cls(address)
        d = unitsentry.info = unit_data
        d['unit_name'] = unit
        d['service'], d['unit'] = unit.split('/')
        unitsentry.upload_scripts()
        return unitsentry

    def upload_scripts(self):
        source = pkg_resources.resource_filename(
            'amulet', os.path.join('unit-scripts', 'amulet'))
        dest = '/tmp/amulet'
        self.run('mkdir -p -m a=rwx {}'.format(dest))
        # copy one at a time b/c `juju scp -r` doesn't work (currently)
        for f in glob.glob(os.path.join(source, '*.py')):
            cmd = "juju scp {} {}:{}".format(
                os.path.join(source, f),
                self.info['unit_name'], dest)
            subprocess.check_call(cmd.split())

    def _fs_data(self, path):
        return self._run_unit_script("filesystem_data.py {}".format(path))

    def file_stat(self, filename):
        return self._fs_data(filename)

    def file_contents(self, filename):
        output, return_code = self._run('cat {}'.format(filename))
        if return_code == 0:
            return output
        else:
            raise IOError(output)

    def directory_stat(self, path):
        return self._fs_data(path)

    def directory_listing(self, path):
        return self._run_unit_script("directory_listing.py {}".format(path))

    def run(self, command):
        output, code = self._run(command)
        return output.strip(), code

    def _run(self, command, unit=None, timeout=300):
        """
        Run a command against an individual unit.

        The timeout defaults to 5m to match the `juju run` command, but can
        be increased to wait for other running hook contexts to complete.
        """
        unit = unit or self.info['unit_name']
        cmd = [
            'juju', 'run',
            '--unit', unit,
            '--timeout', "%ds" % timeout,
            command
        ]
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        output = stdout if p.returncode == 0 else stderr
        return output.decode('utf8'), p.returncode

    def _run_unit_script(self, cmd, timeout=300):
        cmd = "/tmp/amulet/{}".format(cmd)
        output, return_code = self._run(cmd, timeout=timeout)
        if return_code == 0:
            return json.loads(output)
        else:
            raise IOError(output)

    def juju_agent(self, timeout=300):
        return self._run_unit_script("juju_agent.py", timeout)

    def relation(self, from_rel, to_rel):
        this_unit = '{service}/{unit}'.format(**self.info)
        to_service, to_relation = to_rel.split(':')
        r_ids, _ = self._run('relation-ids {}'.format(from_rel))
        r_units = []
        for r_id in r_ids.split():
            r_units.extend(self._run(
                'relation-list -r {}'.format(r_id))[0].split())
        r_units = [u for u in r_units if u.split('/')[0] == to_service]
        for r_unit in r_units:
            r_ids, _ = self._run(
                'relation-ids {}'.format(to_relation), unit=r_unit)
            l_units = []
            for r_id in r_ids.split():
                l_units.extend(self._run(
                    'relation-list -r {}'.format(r_id),
                    unit=r_unit)[0].split())
                if this_unit in l_units:
                    break
            output, _ = self._run(
                'relation-get -r {} - {} --format json'.format(
                    r_id, this_unit), unit=r_unit)
            return json.loads(output)

        raise Exception('Relationship not found')


# Possibly use to build out instead of having code in setup()?
class Talisman(object):
    def __init__(self, services, rel_sentry='relation-sentry', juju_env=None):
        self.unit = {}
        self.service = {}

        if not juju_env:
            juju_env = helpers.default_environment()

        status = self.wait_for_status(juju_env, services)

        for service in services:
            if service not in status['services']:
                continue  # Raise something?

            service_status = status['services'][service]

            if 'units' not in service_status:
                continue

            for unit in service_status['units']:
                unit_data = service_status['units'][unit]
                self.unit[unit] = UnitSentry.fromunitdata(unit, unit_data)
                if 'subordinates' in unit_data:
                    for sub in unit_data['subordinates']:
                        if sub.split('/')[0] not in services:
                            continue
                        subdata = unit_data['subordinates'][sub]
                        self.unit[sub] = UnitSentry.fromunitdata(sub, subdata)

    def __getitem__(self, service):
        """Return the UnitSentry object(s) for ``service``

        :param service: A string in one of two forms::
            "service_name"
            "service_name/unit_num"

        If the first form is used, a list (possibly empty) of all the
        UnitSentry objects for that service is returned.

        If the second form is used, a single object, the UnitSentry for
        the specified unit, is returned.

        Examples::

            >>> d
            <amulet.deployer.Deployment object at 0x7fac83ce8438>
            >>> d.schema()['local']['services']['meteor']['num_units']
            2
            >>> # get UnitSentry for specific unit
            >>> meteor_0 = d.sentry['meteor/0']
            >>> print(meteor_0.info['service'], meteor_0.info['unit'])
            meteor 0
            >>> # get all UnitSentry objects for a service
            >>> for sentry in d.sentry['meteor']:
            ...     print(sentry.info['service'], sentry.info['unit'])
            ...
            meteor 1
            meteor 0
            >>>

        """
        single_unit = '/' in service

        def match(service, unit_name):
            if single_unit:
                return service == unit_name
            return service == unit_name.split('/')[0]

        unit_sentries = [unit_sentry
                         for unit_name, unit_sentry in self.unit.items()
                         if match(service, unit_name)]

        if single_unit and unit_sentries:
            return unit_sentries[0]

        return unit_sentries

    def wait_for_status(self, juju_env, services, timeout=300):
        """Return env status, but only after all units have a
        public-address assigned and are in a 'started' state.

        Some substrates (like Amazon) will return a public-address while the
        machine is still allocating, so it's necessary to also check the
        agent-state to see if the unit is ready.

        Raises if a unit reaches error state, or if public-address not
        available for all units before timeout expires.

        """
        try:
            with helpers.timeout(timeout):
                while True:
                    ready = True
                    status = waiter.status(juju_env)
                    for service in services:
                        if 'units' not in status['services'][service]:
                            continue
                        for unit, unit_dict in \
                                status['services'][service]['units'].items():
                            if 'error' == unit_dict.get('agent-state'):
                                raise Exception('Error on unit {}: {}'.format(
                                    unit, unit_dict.get('agent-state-info')))
                            if 'public-address' not in unit_dict:
                                ready = False
                            if 'started' != unit_dict.get('agent-state'):
                                ready = False
                    if ready:
                        return status
        except helpers.TimeoutError:
            raise helpers.TimeoutError(
                'public-address not set for'
                'all units after {}s'.format(timeout))
        except:
            raise

    def wait(self, timeout=300):
        """
        wait_for_status, called by __init__, blocks until units are in a
        started state. Here we wait for a unit to finish any running hooks.
        When the unit is done, juju_agent will return {}. Otherwise, it returns
        a dict with the name of the running hook.

        Raises an error if the timeout is exceeded.
        """
        ready = False
        try:
            with helpers.timeout(timeout):
                while not ready:
                    for unit in self.unit.keys():
                        status = self.unit[unit].juju_agent(timeout=timeout)

                        # Check if we have a hook key and it's not None
                        if status is None:
                            ready = False
                            break
                        if 'hook' in status and status['hook']:
                            ready = False
                            break
                        else:
                            ready = True
        except:
            raise

    def _sync(self):
        pass


class ServiceSentry(Sentry):
    pass
